/*===========================================================================*/
/*                        Copyright (C) 2009 by                              */
/*                                                                           */
/*                    -- TEC-IT Datenverarbeitung GmbH --                    */
/*   -- team for engineering and consulting in information technologies --   */
/*                                                                           */
/*                           All rights reserved.                            */
/*                                                                           */
/*              This is a part of the TEC-IT Standard Software.              */
/*                                                                           */
/*    This source code is only intended as a supplement to the References    */
/*      and related electronic documentation provided with the product.      */
/*                                                                           */
/*     See these sources for detailed information regarding this product     */
/*===========================================================================*/
package com.tecit.TFPrint;

import java.util.Enumeration;
import java.util.NoSuchElementException;

import com.tecit.TFORMer.Enumerations;
import com.tecit.TFORMer.TFORMer;
import com.tecit.TFORMer.TFormerException;
import com.tecit.TFORMer.Enumerations.EPrinterType;
import com.tecit.TFORMer.Printing.Job;
import com.tecit.TFORMer.Printing.JobData;
import com.tecit.TFORMer.Printing.JobDataCsv;
import com.tecit.TFORMer.Printing.JobDataOdbc;
import com.tecit.TFORMer.Printing.JobDataXml;


/**
 * 
 * <br> <b>Command Line Application (TFPrint)</b> 
 *	<br>
 *	<br>TFPrint offers a stand-alone; command line based printing and reporting application. All required parameters can be specified by command line parameters (e.g. FormLayout to be printed, output device, output type, DataSource, ...). TFPrint is available for all platforms (Microsoft Windows, Linux and UNIX). It can be used in batch-jobs, shell scripts and web-based scripting languages or it can be "spawned" from within your application as separate process. A detailed description is available in a separate manual.
 *
 * @author PhilipK, DavideZ
 *
 */
public class JTFPrint implements TFPrintText
{
	/**
	 * Holds a string that represents the repository or standalone formlayout file
	 */
	private String Repository  =null;
	/**
	 * Holds a string that represents the project of a repository
	 */
	private String Project =null;
	/**
	 * Holds a string that represents the form of a repository
	 */
	private String Form =null;
	/**
	 * String that holds the selected printer or the selected output file
	 */
	private String OutputDevOrFile = null;
	/**
	 * Indicates, if datafield values should be reseted after each record
	 */
	private boolean ClearFieldsOnEachRecord = false;
	/**
	 * Indicates, if datafield values contain escapesequences
	 */
	private boolean TranslateEscapeSequences = false;
	/**
	 * Determines, whether text should be printed (not valid for error texts)
	 */
	private boolean PrintTextWhenExecuted = true;
	/**
	 * Enumeration value for output type, see JNITFormerConstants 
	 */
	private EPrinterType OutputType;
	/**
	 * Holds the path for the selected XSLTfile
	 */
	private String XSLTFile = null;
	/**
	 * Holds the path for the selected jobData file (xml,txt,odbc,...)
	 */
	private String DataFile = null;
	
	private static enum EDataType{
		NONE,XML,TEXT,TFORMER,TFSSCRIPT,ODBC			
	}
	/**
	 * Enumeration value for the type of the datafile
	 */
	private EDataType dataFileType = EDataType.NONE;
	/**
	 * Seperator character for CSV files
	 */
	private char Seperator = ','; 
	/**
	 * Qualifier character for CSV files
	 */
	private char Qualifier = '"';
	/**
	 * ODBC DSN
	 */
	private String DSN = null;
	/**
	 * ODBC user
	 */
	private String USER = null;
	/**
	 * ODBC password
	 */
	private String PWD = null;
	/**
	 * ODBC SQL instruction
	 */
	private String SQL = null;
	/**
	 * Name of the datafield, that holds the number of recordcopies of a repository
	 */
	private String CopyField = null;
	/**
	 * Holds the number of copies to be printed
	 */
	private int PrintCount = 1;
	/**
	 * Name of a Repositories JobTray
	 */
	private String TrayControl = null;
	/**
	 * Holds the startrow
	 */
	private int StartRow = 0;
	/**
	 * holds the startcolumn
	 */
	private int StartCol = 0;
	/**
	 * flag to print the help output
	 */
	private boolean printHelp = false;
	/**
	 * Name of licensee
	 */
	private String licenseeName;
	/**
	 * License key
	 */
	private String licenseKey;
	/**
	 * Number of purchased licenses
	 */
	private int licenseCount;
	/**
	 * Kind of purchased license
	 */
	private Enumerations.ELicenseKind licenseKind;
	/**
	 * Name of the windows printjob
	 */
	private String SpoolJobName = null;
	/**
	 * Holds the path for the config file
	 */
	private String ConfigFile;

	/**
	 * Main method of JTFPrint
	 * 
	 * @param args program arguments
	 * 
	 */
	public static void main(String[] args)
	{
		JTFPrint TFPrint = new JTFPrint();
		boolean res = true;

		// no parameter
		if( args.length==0 ){
			System.out.println(sStartText);
			System.exit(255);
		}
		
		// read and verify the arguments
		res = TFPrint.parseArgs(args);
		if( !res ){
			System.out.println("Please try 'JTFPrint -h' to display the commandline help of JTFPrint.");
			System.exit(1);
		}
		
		// execute the print command
		try {
			if( TFPrint.printHelp ){
				System.out.println(sStartText);
				System.out.println(sHelpText);
			} else {
				// print useful informations about the system
				if( TFPrint.PrintTextWhenExecuted )
					TFPrint.printExecutionMessage();
				
				res = TFPrint.executeTFPrint();
			}
		} 
		catch (Throwable ex){
			if( ex instanceof UnsatisfiedLinkError ){
				// Exception due to unknown library
				TFPrint.error("Unexpected error", ex);
				System.err.print  ("You must include the TFORMer library into your Java library path (");
				System.err.print  (System.getProperty("java.library.path"));
				System.err.println(")");
			}
			else if( ex instanceof TFormerException ){
				// TFormerException
				TFormerException tfex = (TFormerException) ex;
				System.err.println(tfex.getMessage());
				if( tfex.getErrorCode()!=TFormerException.ERR_OK ){
					System.err.println(tfex.getInformation());
					System.err.println("Source: "+tfex.getSourceReference());
					System.err.println("Error code: "+tfex.getWin32ErrorCode());
				}
				tfex.printStackTrace();
			} else {
				// generic exception 
				TFPrint.error("Unexpected error", ex);
			}
		}
		
		// return an error 
		if( !res )
			System.exit(1);
	}
	
	/**
	 * Print useful informations about the system
	 */
	public void printExecutionMessage()
	{
		System.out.println("\n----------------Java TFPrint-------------------\n");
		System.out.println("Java version: " + System.getProperty("java.version"));
		System.out.println("Library path: " + System.getProperty("java.library.path"));
		System.out.println("OS name     : " + System.getProperty("os.name"));
	}
	
	/**
	 * Write an error
   * @param msg error message 
   * @param ex exception occurred
	 */
	public void error(String msg, Throwable ex)
	{
		System.err.println(msg);
		if( ex!=null )
			ex.printStackTrace();
	}

	/**
	 * Write an error
   * @param msg error message 
	 */
	public void error(String msg)
	{
		error(msg, null);
	}
	
	/**
	 * Iterate through the arguments
	 * @param args argument array from main
	 * @return true if succeeded
	 * @throws TFormerException 
	 */
	public boolean parseArgs(String[] args)
	{
		return init(new ParseArgument(args));
	}
	
	
	/**
	 * Initializes the instance with the parameters
	 * @param parse instance of the parser that reads the parameters 
	 * @return true if succeeded
	 */
	public boolean init(ParseArgument parse)
	{
		String command=null, param=null, option; 

		try {
			while( parse.hasMoreElements() ){
				// find the next parameter, generating an exception if doesn't exist
				command = parse.next(false);
	
				if( command.equalsIgnoreCase("-H") 
				 || command.equalsIgnoreCase("-?") 
				 || command.equalsIgnoreCase("-HELP") 
				 || command.equalsIgnoreCase("--HELP") ){  
					//
					// HELP command
					//
					this.printHelp = true;
					return true;
				}
				else if( command.equalsIgnoreCase("-LCUPS")){
					//
					// LCUPS command
					//
					String osName = System.getProperty("os.name");
					if(osName.startsWith("Linux")){	
						// Linux support
						error("Cups is not yet implemented");
						return false;
					}
					else if(osName.startsWith("Windows")){
						// Windows no supports CUPS
						error("LCUPS not supported on Windows plattform");
						return false;
					} else {
						// Others OS
						error("LCUPS not supported on plattform "+osName);
						return false;
					}
				}
				else if( command.equalsIgnoreCase("-SYSTEMID")){
					//
					// SYSTEMID command
					//
					error("SystemID not supported");
					return false;
				}
				else if( command.equalsIgnoreCase("-Q")){
					//
					// Quiet command 
					//
					PrintTextWhenExecuted = false;
				}
				else if( command.equalsIgnoreCase("-L")){
					//
					// License Key command
					//
					if( !parse.isNextCommand() )
						this.licenseeName = parse.next(false);
					if( !parse.isNextCommand() )
						this.licenseKey = parse.next(false);
					if( !parse.isNextCommand() )
						this.licenseCount = Integer.parseInt(parse.next(false));
					param = (parse.isNextCommand() ? null : parse.next(false));
					if( param==null ){
						// we ignore this, as in tfprint
					}
					else if( param.equalsIgnoreCase("Single") ){
						this.licenseKind = Enumerations.ELicenseKind.Workstation;
					}
					else if( param.equalsIgnoreCase("Site") ){
						this.licenseKind = Enumerations.ELicenseKind.Site;
					}
					else if( param.equalsIgnoreCase("Developer") ){
						this.licenseKind = Enumerations.ELicenseKind.Developer;
					}
					else if( param.equalsIgnoreCase("Office") ){
						this.licenseKind = Enumerations.ELicenseKind.Office;
					}
					else if( param.equalsIgnoreCase("Workgroup") ){
						this.licenseKind = Enumerations.ELicenseKind.Workgroup;
					}
					else if( param.equalsIgnoreCase("Server") ){
						this.licenseKind = Enumerations.ELicenseKind.Server;
					} else {
						error("Invalid argument '" + param + "' for parameter '" + command + "'");
						return false;
					}
				}
				else if( command.equalsIgnoreCase("-F")){				
					//
					// Form specification command
					//
					Repository = parse.next(false);
					if( !parse.isNextCommand() )
						Project = parse.next(false);
					if( !parse.isNextCommand() )
						Form  = parse.next(false);
				}
				else if( command.equalsIgnoreCase("-CONFIG")){
					//
					// Form specification command
					//
					ConfigFile = parse.next(false);
					
					error("ConfigFile not supported");
					return false;
				}
				else if( command.equalsIgnoreCase("-C")){
					//
					// CommandFileASCII command
					//
					param = parse.next(false);
					
					error("Commandfile not supported");
					return false;
				}
				else if( command.equalsIgnoreCase("-D")){
					//
					// Data File command
					//
					DataFile = parse.next(false);
				}
				else if( command.equalsIgnoreCase("-DT")){
					//
					// Data File Type command
					//
					param = parse.next(false);
					if(param.equalsIgnoreCase("XML")){
						dataFileType = EDataType.XML;
					}
					else if(param.equalsIgnoreCase("TEXT")){
						dataFileType = EDataType.TEXT;
					}
					else if(param.equalsIgnoreCase("TFORMER")){
						dataFileType = EDataType.TFORMER;
					}
					else if(param.equalsIgnoreCase("TFSSCRIPT") || param.equalsIgnoreCase("TFS")){
						dataFileType = EDataType.TFSSCRIPT;
					}
					else if(param.equalsIgnoreCase("ODBC")){
						dataFileType = EDataType.ODBC;
					} else {
						error("Invalid argument '" + param + "' for parameter '" + command + "'");
						return false;
					}
				}
				else if( command.equalsIgnoreCase("-DO") ){
					//
					// Data File Input Data Options command
					//
					// create a loop to read all parameters unless a new command
					do {
						param = parse.next(false);
						if( param.startsWith("CL") && param.length()>2 ){
							// Clear data fields between records
							option = param.substring(2);
							if( option.equals("yes") || option.equals("y") || option.equals("1") ){
								ClearFieldsOnEachRecord = true;
							}
							else if( option.equals("no") || option.equals("n") || option.equals("0") ){
								ClearFieldsOnEachRecord = false;
							} else {
								error("Invalid argument '" + param + "' for parameter '" + command + "'");
								return false;
							}
						}
						else if( param.startsWith("SC") && param.length()==3 ){
							// Separator character
							Seperator = param.charAt(3);
						}
						else if( param.startsWith("QC") && param.length()==3 ){
							// Qualificator character
							Qualifier = param.charAt(3);
						}
						else if( param.startsWith("ESC") && param.length()>3 ){
							// Escape Sequences
							option = param.substring(3);
							if( option.equals("yes") || option.equals("y") || option.equals("1") ){
								TranslateEscapeSequences = true;
							}
							else if( option.equals("no") || option.equals("n") || option.equals("0") ){
								TranslateEscapeSequences = false;
							} else {
								error("Invalid argument '" + param + "' for parameter '" + command + "'");
								return false;
							}
						}
						else if( param.startsWith("RCA") && param.length()>3 ){
							// Attribute specifying record copy counter
							
							error("Invalid argument '" + param + "' for parameter '" + command + "' not supported");
							return false;
						}
						else if( param.startsWith("RCF") && param.length()>3 ){
							// Fieldname specifying record copy counter
							CopyField = param.substring(3);
						  // Currently not possible using wrapper *
							//CopyFieldIsAttribute = false;
						}
						else if( param.startsWith("DSN:") && param.length()>4 ){
							DSN = param.substring(4);
						}
						else if( param.startsWith("PWD:") && param.length()>4 ){
							PWD = param.substring(4);
						}
						else if( param.startsWith("SQL:") && param.length()>4 ){
							SQL = param.substring(4);
						}
						else if( param.startsWith("USER:") && param.length()>5 ){
							USER = param.substring(5);
						}
						else if( param.startsWith("XSLT:") && param.length()>5 ){
							XSLTFile = param.substring(5);
						} else { 
							error("Invalid parameter '" + param + "'");
							return false;
						}
					}
					while( !parse.isNextCommand() );
				}
				else if( command.equalsIgnoreCase("-O")){
					//
					// Output Device Name command
					//
					OutputDevOrFile = parse.next(false);
				}
				else if( command.equalsIgnoreCase("-OT")){
					//
					// Output Device Type command
					//
					param = parse.next(false);
					if(param.equalsIgnoreCase("WIN32")){
						OutputType = EPrinterType.Win32;}
					else if(param.equalsIgnoreCase("CUPS")){
						OutputType = EPrinterType.Cups;}
					else if(param.equalsIgnoreCase("REPO")){
						OutputType = EPrinterType.Repository;}
					else if(param.equalsIgnoreCase("PS")){
						OutputType = EPrinterType.PSFile;}
					else if(param.equalsIgnoreCase("PSPRINTER")){
						OutputType = EPrinterType.PSPrinter;}
					else if(param.equalsIgnoreCase("PDF")){
						OutputType = EPrinterType.PDFFile;}
					else if(param.equalsIgnoreCase("HTML")){
						OutputType = EPrinterType.HTML;}
					else if(param.equalsIgnoreCase("TXT")){
						OutputType = EPrinterType.TxtFile;}
					else if(param.equalsIgnoreCase("TXTPRINTER")){
						OutputType = EPrinterType.TxtPrinter;}
					else if(param.equalsIgnoreCase("IMGBMP")){
						OutputType = EPrinterType.ImageBmp;}
					else if(param.equalsIgnoreCase("IMGGIF")){
						OutputType = EPrinterType.ImageGif;}
					else if(param.equalsIgnoreCase("IMGJPG")){
						OutputType = EPrinterType.ImageJpg;}
					else if(param.equalsIgnoreCase("IMGPCX")){
						OutputType = EPrinterType.ImagePcx;}
					else if(param.equalsIgnoreCase("IMGPNG")){
						OutputType = EPrinterType.ImagePng;}
					else if(param.equalsIgnoreCase("IMGTGA")){
						OutputType = EPrinterType.ImageTga;}
					else if(param.equalsIgnoreCase("IMGTIF")){
						OutputType = EPrinterType.ImageTifSingle;}
					else if(param.equalsIgnoreCase("IMGMULTITIF")){
						OutputType = EPrinterType.ImageTifMultiPage;}
					else if(param.equalsIgnoreCase("ZPL")){
						OutputType = EPrinterType.ZPLFile;}
					else if(param.equalsIgnoreCase("ZPLPRINTER")){
						OutputType = EPrinterType.ZPLPrinter;}
					else {
						error("Invalid argument '" + param + "' for parameter '" + command + "'");
						return false;
					}
				}
				else if( command.equalsIgnoreCase("-OO")){
					//
					// Output Options command
					//
					// create a loop to read all parameters unless a new command
					do {
						param = parse.next(false);
						if( param.startsWith("CP") && param.length()>2 ){
							PrintCount = Integer.parseInt(param.substring(2));
						}
						else if( param.startsWith("SR") && param.length()>2 ){
							StartRow = Integer.parseInt(param.substring(2));
						}
						else if( param.startsWith("SC") && param.length()>2 ){
							StartCol = Integer.parseInt(param.substring(2));
						}
						else if( param.startsWith("TC") && param.length()>2 ){
							TrayControl = param.substring(2);
						}
						else if( param.startsWith("JT") && param.length()>2 ){
							SpoolJobName = param.substring(2);
						}
						else if( param.startsWith("DXN") ){
							error("Argument '" + param + "' for parameter '" + command + "' not supported");
							return false;
						}
						else if( param.startsWith("DXH") ){
							error("Argument '" + param + "' for parameter '" + command + "' not supported");
							return false;
						}
						else if( param.startsWith("DXV") ){
							error("Argument '" + param + "' for parameter '" + command + "' not supported");
							return false;
						}
						else {
							error("Invalid argument '" + param + "' for parameter '" + command + "'");
							return false;
						}
					}
					while( !parse.isNextCommand() );
				} else{
					error("Invalid parameter " + command);
					return false;
				}
			}
		}
		catch(NumberFormatException ex){
			// exception occurs when convert string to number in a parameter
			error("Invalid argument '" + param + "' for parameter '" + command + "'");
			return false;
		}
		catch(NoSuchElementException ex){
			// exception occurs when parse has finished the parameters
			error("Missing arguments for parameter '" + command + "'");
			return false;
		}
		return true;	
	}
	

	/**
	 * Executes TFormer with previous set properties
	 * @return true if succeeded
	 * @throws TFormerException if an error occured. Errors can be as follows: 
	 * 
	 * <br>					==> A given value from the commandline is invalid
	 * <br>					==> A Repository or standalone-FormLayout cannot be opened
	 * <br>					==> The extension of a file is not supported
	 * <br>					==> ...
	 */
	public boolean executeTFPrint()
	throws TFormerException
	{
		Job job = new Job();
		JobData jobData = null;

		try {
			// set the license (if it exists)
			if( this.licenseeName!=null ){
				TFORMer.license(this.licenseeName, this.licenseKind
						           ,this.licenseCount, this.licenseKey);
			}
			
			// setting the repository
			if( Repository==null ){
				error("No form layout specified. Please specify form layout using parameter -F!");
				return false;
			}
			else if( Repository.endsWith(".tff") ){
				// StandAloneFormLayout
				job.setRepositoryName(Repository);
			} 
			else if( Repository.endsWith(".tfr") ){
				// Repository-based FormLayout
				job.setRepositoryName(Repository);
				job.setProjectName(Project);
				job.setFormName(Form);
			} else{
				error("Invalid file extension of the repository file (-F)");
				return false;
			}

			// setting the configuration file
			if(ConfigFile != null){
				job.setConfigurationFile(ConfigFile);
			}

			// setting the type of data file
			switch(dataFileType)
			{
				case NONE:
					break;
					
				case ODBC:
					if(DSN==null || SQL==null ){
						error("No DSN or SQL-statement specified. Please specify data type using parameter -DT!");
						return false;
					}
					jobData = new JobDataOdbc(DSN, USER, PWD, SQL);
					break;
					
				case XML:
					jobData = new JobDataXml(DataFile);
					break;
					
				case TEXT:
					if(DataFile==null ){
						error("No DataFile specified. Please specify data type using parameter -D!");
						return false;
					}
					jobData = new JobDataCsv(DataFile,Seperator,Qualifier);
					break;
					
				case TFORMER:
					if(DataFile==null ){
						error("No DataFile specified. Please specify data type using parameter -D!");
						return false;
					}
					jobData = new JobDataCsv(DataFile,Seperator,Qualifier);
					break;
					
				case TFSSCRIPT: 
					error("TFormerServerScript not supported in JTFPrint!");
					return false;
					
				default:
					error("Invalid Datatype specified!");
					return false;
			}

			// setting for data
			if( jobData!=null ){
				jobData.setHandleEscapeSequences(TranslateEscapeSequences);
				jobData.setReuseValues(!ClearFieldsOnEachRecord);
				job.setJobData(jobData);
			}
			
			if(TrayControl != null)
				job.setJobTrayControlName(TrayControl);
			
			job.setStartColumn(StartCol);
			job.setStartRow(StartRow);
		
			//When the JTFPrint parameters are right, a princommand will be executed.
			//If the -O command was not executed, the form will be printed on the defaultprinter
			job.setNumberOfCopies(PrintCount);
			
			if(		
					OutputType == EPrinterType.PSFile || OutputType == EPrinterType.PDFFile || OutputType == EPrinterType.HTML ||
					OutputType == EPrinterType.TxtFile 	|| OutputType == EPrinterType.ImageBmp ||
					OutputType == EPrinterType.ImageGif 	|| OutputType == EPrinterType.ImageJpg ||
					OutputType == EPrinterType.ImagePcx 	|| OutputType == EPrinterType.ImagePng ||
					OutputType == EPrinterType.ImageTga 	|| OutputType == EPrinterType.ImageTifSingle ||
					OutputType == EPrinterType.ImageTifMultiPage 	|| OutputType == EPrinterType.ZPLFile){
				
				job.setPrinterType((OutputType));
				
				job.setOutputName(OutputDevOrFile);
				
				job.setPrinterName(null);
				
			}
			else {
				job.setOutputName(SpoolJobName);
				
				job.setPrinterName(OutputDevOrFile);
			}

			if(PrintTextWhenExecuted){
				// see tfprint output
				System.out.println("Repository  : " + Repository);
				System.out.println("Project     : " + Project);
				System.out.println("Form        : " + Form);
				System.out.println("Output      : " + OutputDevOrFile);
				System.out.println();
			}
			
			// generate the print
			job.print();
		}
		catch(TFormerException ex){
			throw ex;
		}
		finally {
			// dispose the resources
			job.dispose();
			job = null;
		}
		
		if(PrintTextWhenExecuted){
			// see tfprint output
			System.out.println();
			System.out.println("Print job finished.");
		}
		return true;
	}
	/**
	 * Simple class to parse argument
	 * @author davidez
	 */
	protected class ParseArgument implements Enumeration
	{
		/** array of the parameters */
		private String[] args;
		/** index of the last parameter read */
		private int idx;
		
		/**
		 * Constructor
		 * @param args array of the parameters
		 */
		public ParseArgument(String[] args)
		{
			this.args = args;
			this.idx = -1;
		}
		
		/**
		 * Return the number of the parameter
		 */
		public int lenght()
		{
			if( this.args==null )
				return 0;
			return this.args.length;
		}

		/**
		 * Test if the next parameter is a command (string starts with 
		 * <code>-</code> character
		 */
		public boolean isNextCommand()
		{
			if( !hasMoreElements() )
				return true;
			
			return this.args[this.idx+1].startsWith("-");
		}
		
		/**
		 * Return the next parameter
		 * @param ignoreException flag to ignore the exception if there isn't a new
		 *                        parameter 
		 * @return <code>null</code> if no exists other parameter and the exception
		 *         is ignored, otherwise a valid string 
		 * @throws NoSuchElementException exception if no exists other parameter
		 */
		public String next(boolean ignoreException)
		throws NoSuchElementException
		{
			if( hasMoreElements() )
				return this.args[++this.idx];
			
			if( ignoreException )
				return null;

			throw new NoSuchElementException();
		}
		
		/**
		 * Test if exists a next element
		 */
		public boolean hasMoreElements()
		{
			return (this.args!=null && this.idx+1<this.args.length);
		}
		
		/**
		 * Return the next element
		 */
		public Object nextElement()
		throws NoSuchElementException
		{
			return next(false);
		}
	}
}
